; File......: PUTKEY.ASM
; Author....: Ted Means
; Date......: $Date:   15 Aug 1991 23:07:10  $
; Revision..: $Revision:   1.2  $
; Log file..: $Logfile:   E:/nanfor/src/putkey.asv  $
; 
; This is an original work by Ted Means and is placed in the
; public domain.
;
; Modification history:
; ---------------------
;
; $Log:   E:/nanfor/src/putkey.asv  $
;  
;     Rev 1.2   15 Aug 1991 23:07:10   GLENN
;  Forest Belt proofread/edited/cleaned up doc
;  
;     Rev 1.1   14 Jun 1991 19:54:56   GLENN
;  Minor edit to file header
;  
;     Rev 1.0   01 Apr 1991 01:03:48   GLENN
;  Nanforum Toolkit
;  

; $DOC$
; $FUNCNAME$
;     FT_PUTKEY()
; $CATEGORY$
;     Keyboard/Mouse
; $ONELINER$
;     Stuff a keystroke into the keyboard buffer
; $SYNTAX$
;     FT_PUTKEY( <nKeyValue> ) -> lResult
; $ARGUMENTS$
;     <nKeyValue> is the INKEY() value of the keystroke to be stuffed.
; $RETURNS$
;    .T. if the keystroke was put into the keyboard buffer.
;    .F. if nKeyValue was invalid or the buffer was full.
; $DESCRIPTION$
;    This function is similar to the KEYBOARD command, with a few
;    exceptions. First, this function does not clear the keyboard buffer
;    before inserting the keystroke.  In addition, since it uses the
;    Inkey() value, you can stuff any key, including function keys, into
;    the keyboard buffer. However, this also means that unlike the KEYBOARD
;    command, you can only stuff one keystroke at a time.
;
;    You can easily create a User-Defined Command that makes this function
;    even more like the KEYBOARD command.  For example,
;
;        #command KEYSTROKE <key> => FT_PUTKEY( <key> )
;
;    will create a command called KEYSTROKE that could be used as a
;    companion command to KEYBOARD.  The only difference is that it would
;    insert a single keystroke instead of a string.
;
;    Be aware that early releases of Clipper 5.0 have a bug in their
;    keyboard handling.  Among other things, Set Typeahead does not work
;    correctly, so this function may at times return .T. even though the
;    keyboard buffer is full.  The best solution is not to attempt stuffing
;    more than 15 keys at a time into the keyboard buffer.
;
;    This function is written to adhere to Turbo Assembler's IDEAL mode.
;    To use another assembler, rearrange the SEGMENT and PROC directives
;    and make any other necessary changes to the source code.
; $EXAMPLES$
;      FT_PUTKEY( -9 )   //  Stuff the F10 key
;      FT_PUTKEY( 276 )  //  Stuff the Alt T key
;      KEYSTROKE 28      //  Stuff the F1 key using a User-Defined Command
; $END$


         IDEAL

Public   FT_PUTKEY

EXTRN    __PARINFO:FAR
EXTRN    __PARNI:FAR
EXTRN    __RETL:FAR

SEGMENT  _NANFOR   WORD      PUBLIC    "CODE"
         ASSUME    CS:_NANFOR

PROC     FT_PUTKEY  FAR

         XOR       AX,AX                     ; PREPARE TO COUNT PARAMS
         PUSH      AX                        ; SAVE ON STACK
         CALL      __PARINFO                 ; GET PARAM COUNT
         ADD       SP,2                      ; REALIGN STACK
         OR        AX,AX                     ; ZERO PARAMS?
         JNZ       TEST1                     ; IF NOT, CONTINUE
         JMP       DONE                      ; RETURN VALUE = FALSE, GO TO END

TEST1:   MOV       AX,1                      ; PREPARE TO CHECK PARAMETER #1
         PUSH      AX                        ; SAVE PARAMETER # ON STACK
         CALL      __PARINFO                 ; CALL PARAMETER INFO ROUTINE
         ADD       SP,2                      ; REALIGN STACK
         TEST      AX,2                      ; IS PARAMETER NUMERIC?
         JNZ       GET1                      ; IF SO, CONTINUE
BADPARAM:XOR       AX,AX                     ; SET RETURN VALUE TO FALSE
         JMP       DONE                      ; GO TO END

GET1:    MOV       AX,1                      ; PREPARE TO RETRIEVE PARAMETER #1
         PUSH      AX                        ; SAVE PARAMETER # ON STACK
         CALL      __PARNI                   ; RETRIEVE PARAMETER
         ADD       SP,2                      ; REALIGN STACK

         CMP       AX,385                    ; TEST HIGHEST INKEY()
         JG        BadParam                  ; Bad INKEY() value
         Cmp       AX,-39                    ; Test lowest INKEY()
         JL        BadParam                  ; Process error

CtrlF1:  Mov       CL,0                      ; Set ASCII value to null
         Cmp       AX,-10                    ; Is Ctrl F1 thru Alt F10?
         JG        F2                        ; If not, check next range
         Neg       AX                        ; Get absolute value of AX
         Add       AL,74                     ; Translate INKEY() to scan code
         Mov       CH,AL                     ; Move scan code to CH
         JMP       StuffIt                   ; Stuff the keystroke

F2:      Or        AX,AX                     ; See if key is F2 thru F10
         JNS       F1                        ; If not, check next range
         Neg       AX                        ; Get absolute value of AX
         Add       AL,59                     ; Translate INKEY() to scan code
         Mov       CH,AL                     ; Move scan code to CH
         JMP       StuffIt                   ; Stuff the keystroke

F1:      Cmp       AX,28                     ; See if key is F1
         JNE       CtrlF                     ; If not, check next key
         Mov       CH,59                     ; Supply scan code for F1
         JMP       StuffIt                   ; Stuff the keystroke

CtrlF:   Cmp       AX,6                      ; See if key is Ctrl F or End
         JNE       CtrlW                     ; If not, check next key
         Mov       CH,79                     ; Supply scan code for End
         JMP       StuffIt                   ; Stuff the keystroke

CtrlW:   Cmp       AX,23                     ; See if key is Ctrl W or Ctrl End
         JNE       CtrlHome                  ; If not, check next key
         Mov       CH,117                    ; Supply scan code for Ctrl End
         JMP       Short StuffIt             ; Stuff the keystroke

CtrlHome:Cmp       AX,29                     ; See if key is Ctrl Home or Ctrl]
         JNE       CtrlV                     ; If not, check next key
         Mov       CH,119                    ; Supply scan code for Ctrl Home
         JMP       Short StuffIt             ; Stuff the keystroke

CtrlV:   Cmp       AX,22                     ; See if key is Ins or Ctrl V
         JNE       ShiftTab                  ; If not, check next key
         Mov       CH,82                     ; Supply scan code for Ctrl V
         JMP       Short StuffIt             ; Stuff the keystroke

ShiftTab:Cmp       AX,271                    ; See if key is Shift Tab
         JNE       CtrlPgDn                  ; If not, check next key
         Mov       CH,15                     ; Supply scan code for Shift Tab
         JMP       Short StuffIt             ; Stuff the keystroke

CtrlPgDn:Cmp       AX,30                     ; See if key is Ctrl PgDn
         JNE       CtrlPgUp                  ; If not, check next key
         Mov       CH,118                    ; Supply scan code for Ctrl PgDn
         JMP       Short StuffIt             ; Stuff the keystroke

CtrlPgUp:Cmp       AX,31                     ; See if key is Ctrl PgUp
         JNE       AltQ                      ; If not, check next key
         Mov       CH,132                    ; Supply scan code for Ctrl PgUp
         JMP       Short StuffIt             ; Stuff the keystroke

AltQ:    Cmp       AX,272                    ; See if key is Alt Q . . .
         JL        ASCII
         Cmp       AX,281                    ; . . . thru Alt P
         JG        AltA
         Mov       CH,AL
         JMP       Short StuffIt

AltA:    Cmp       AX,286                    ; See if key is Alt A . . .
         JNL       AltL
         JMP       BadParam
AltL:    Cmp       AX,294                    ; . . . thru Alt L
         JG        AltZ
         Mov       CH,AL
         JMP       Short StuffIt

AltZ:    Cmp       AX,300                    ; See if key is Alt Z . . .
         JNL       AltM
         JMP       BadParam
AltM:    Cmp       AX,306                    ; . . . thru Alt M
         JG        Alt1
         Mov       CH,AL
         Mov       CL,0
         JMP       Short StuffIt

Alt1:    Cmp       AX,376                    ; See if key is Alt 1 . . .
         JNL       AltNum
         JMP       BadParam
AltNum:  Mov       CH,AL
         Mov       CL,0
         JMP       Short StuffIt

ASCII:   Or        AH,AH                     ; See if key is plain ASCII
         JZ        Okay
         JMP       BadParam
Okay:    Mov       CX,AX

StuffIt: CLI                                 ; Disable keyboard temporarily
         Push      DS                        ; Save DS
         Xor       AX,AX                     ; Clear AX
         Mov       DS,AX                     ; Point DS to low RAM
         Mov       AX,[Word Ptr 41Ch]        ; Get tail pointer
         Add       AX,2                      ; Make room for keystroke
         Cmp       AX,3Eh                    ; Need to wrap?
         JNE       IsItFull                  ; No, so continue
         Mov       AX,1Eh                    ; Wrap tail
IsItFull:Cmp       AX,[Word Ptr 41Ah]        ; Is buffer full?
         JNE       PutChar                   ; No, so continue
         Xor       AX,AX                     ; Set return value to .F.
         Jmp       Short Exit                ; Quit

PutChar: Mov       BX,[Word Ptr 41Ch]        ; Get tail pointer
         Add       BX,400h                   ; Adjust offset by constant
         Mov       [Word Ptr BX],CX          ; Stuff keystroke
         Mov       [Word Ptr 41Ch],AX        ; Save new buffer tail
         Mov       AX,1                      ; Set return value to .T.

Exit:    STI                                 ; Enable keystrokes again
         Pop       DS                        ; Restore DS

Done:    Push      AX                        ; Save return value on stack
         Call      __RetL                    ; Return it to Clipper app
         Add       SP,2                      ; Realign stack
         Ret
Endp     FT_PutKey
Ends     _NanFor
End
